"""Test Python APIs for target, breakpoint, and process."""

import os, sys, time
import unittest2
import lldb
from lldbtest import *

class TestHelloWorld(TestBase):

    mydir = "hello_world"

    @unittest2.skipUnless(sys.platform.startswith("darwin"), "requires Darwin")
    def test_with_dsym_and_run_command(self):
        """Create target, breakpoint, launch a process, and then kill it.

        Use dsym info and lldb "run" command.
        """
        self.buildDsym()
        self.hello_world_python(useLaunchAPI = False)

    @unittest2.expectedFailure
    def test_with_dwarf_and_process_launch_api(self):
        """Create target, breakpoint, launch a process, and then kill it.

        Use dwarf map (no dsym) and process launch API.
        """
        self.buildDwarf()
        self.hello_world_python(useLaunchAPI = True)

    def hello_world_python(self, useLaunchAPI):
        """Create target, breakpoint, launch a process, and then kill it."""

        exe = os.path.join(os.getcwd(), "a.out")

        target = self.dbg.CreateTarget(exe)

        breakpoint = target.BreakpointCreateByLocation("main.c", 4)

        # The default state after breakpoint creation should be enabled.
        self.assertTrue(breakpoint.IsEnabled(),
                        "Breakpoint should be enabled after creation")

        breakpoint.SetEnabled(False)
        self.assertTrue(not breakpoint.IsEnabled(),
                        "Breakpoint.SetEnabled(False) works")

        breakpoint.SetEnabled(True)
        self.assertTrue(breakpoint.IsEnabled(),
                        "Breakpoint.SetEnabled(True) works")

        # rdar://problem/8364687
        # SBTarget.LaunchProcess() issue (or is there some race condition)?

        if useLaunchAPI:
            # The following approach of launching a process looks untidy and only
            # works sometimes.
            process = target.LaunchProcess([''], [''], os.ctermid(), False)

            SR = process.GetThreadAtIndex(0).GetStopReason()
            count = 0
            while SR == StopReasonEnum("Invalid") or SR == StopReasonEnum("Signal"):
                print >> sys.stderr, "StopReason =", StopReasonString(SR)

                time.sleep(1.0)
                print >> sys.stderr, "Continuing the process:", process
                process.Continue()

                count = count + 1
                if count == 10:
                    print >> sys.stderr, "Reached 10 iterations, giving up..."
                    break

                SR = process.GetThreadAtIndex(0).GetStopReason()
        else:
            # On the other hand, the following two lines of code are more reliable.
            self.runCmd("run")
            process = target.GetProcess()

        self.runCmd("thread backtrace")
        self.runCmd("breakpoint list")
        self.runCmd("thread list")

        # The stop reason of the thread should be breakpoint.
        thread = process.GetThreadAtIndex(0)
        
        print >> sys.stderr, "StopReason =", StopReasonString(thread.GetStopReason())
        self.assertTrue(thread.GetStopReason() == StopReasonEnum("Breakpoint"),
                        STOPPED_DUE_TO_BREAKPOINT)

        # The breakpoint should have a hit count of 1.
        self.assertTrue(breakpoint.GetHitCount() == 1, BREAKPOINT_HIT_ONCE)

        # Now kill the process, and we are done.
        rc = process.Kill()
        self.assertTrue(rc.Success())


if __name__ == '__main__':
    import atexit
    lldb.SBDebugger.Initialize()
    atexit.register(lambda: lldb.SBDebugger.Terminate())
    unittest2.main()
